@@ -1,5 +1,5 @@
 /* stty -- change and print terminal line settings
-   Copyright (C) 1990, 1991 Free Software Foundation, Inc.
+   Copyright (C) 90, 91, 92, 93, 1994 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -58,7 +58,7 @@
 #endif
 
 #include "system.h"
-#include "version.h"
+#include "long-options.h"
 
 #if defined(GWINSZ_BROKEN)	/* Such as for SCO UNIX 3.2.2. */
 #undef TIOCGWINSZ
@@ -103,6 +103,14 @@
 #if defined(VSWTCH) && !defined(CSWTCH)
 #define	CSWTCH _POSIX_VDISABLE
 #endif
+
+/* SunOS 5.3 loses (^Z doesn't work) if `swtch' is the same as `susp'.
+   So the default is to disable `swtch.'  */
+#if defined (__sparc__) && defined (__svr4__)
+#undef CSWTCH
+#define	CSWTCH _POSIX_VDISABLE
+#endif
+
 #if defined(VDSUSP) && !defined (CDSUSP)
 #define	CDSUSP Control ('y')
 #endif
@@ -118,35 +126,36 @@
 #if defined(VLNEXT) && !defined(CLNEXT)
 #define	CLNEXT Control ('v')
 #endif
-#if defined(VFLUSHO) && !defined(CFLUSHO)
-#define CFLUSHO Control ('o')
-#endif
 #if defined(VDISCARD) && !defined(VFLUSHO)
 #define VFLUSHO VDISCARD
 #endif
+#if defined(VFLUSHO) && !defined(CFLUSHO)
+#define CFLUSHO Control ('o')
+#endif
 #if defined(VSTATUS) && !defined(CSTATUS)
 #define CSTATUS Control ('t')
 #endif
 
-char *visible ();
-unsigned long baud_to_value ();
-int recover_mode ();
-int screen_columns ();
-int set_mode ();
-long integer_arg ();
-speed_t string_to_baud ();
-tcflag_t *mode_type_flag ();
-void display_all ();
-void display_changed ();
-void display_recoverable ();
-void display_settings ();
-void display_speed ();
-void display_window_size ();
+static char *visible ();
+static unsigned long baud_to_value ();
+static int recover_mode ();
+static int screen_columns ();
+static int set_mode ();
+static long integer_arg ();
+static speed_t string_to_baud ();
+static tcflag_t *mode_type_flag ();
+static void display_all ();
+static void display_changed ();
+static void display_recoverable ();
+static void display_settings ();
+static void display_speed ();
+static void display_window_size ();
+static void sane_mode ();
+static void set_control_char ();
+static void set_speed ();
+static void set_window_size ();
+
 void error ();
-void sane_mode ();
-void set_control_char ();
-void set_speed ();
-void set_window_size ();
 
 /* Which speeds to set.  */
 enum speed_setting
@@ -388,18 +397,10 @@ static int max_col;
 /* Current position, to know when to wrap. */
 static int current_col;
 
-/* If non-zero, display usage information and exit.  */
-static int show_help;
-
-/* If non-zero, print the version on standard output and exit.  */
-static int show_version;
-
 static struct option longopts[] =
 {
   {"all", no_argument, NULL, 'a'},
-  {"help", no_argument, &show_help, 1},
   {"save", no_argument, NULL, 'g'},
-  {"version", no_argument, &show_version, 1},
   {NULL, 0, NULL, 0}
 };
 
@@ -411,7 +412,7 @@ char *program_name;
    Print a space first unless MESSAGE will start a new line. */
 
 /* VARARGS */
-void
+static void
 #ifdef __STDC__
 wrapf (char *message, ...)
 #else
@@ -428,7 +429,7 @@ wrapf (message, va_alist)
   vsprintf (buf, message, args);
   va_end (args);
   buflen = strlen (buf);
-  if (current_col + buflen >= max_col)
+  if (current_col + (current_col > 0) + buflen >= max_col)
     {
       putchar ('\n');
       current_col = 0;
@@ -451,7 +452,11 @@ usage (status)
 	     program_name);
   else
     {
-      printf ("Usage: %s [OPTION]... [SETTING]...\n", program_name);
+      printf ("\
+Usage: %s [SETTING]...\n\
+  or:  %s OPTION\n\
+",
+	      program_name, program_name);
       printf ("\
 \n\
   -a, --all       print all current settings in human-readable form\n\
@@ -459,7 +464,7 @@ usage (status)
       --help      display this help and exit\n\
       --version   output version information and exit\n\
 \n\
-Optionnal - before SETTING indicates negation.  An * marks non-POSIX\n\
+Optional - before SETTING indicates negation.  An * marks non-POSIX\n\
 settings.  The underlying system defines which settings are available.\n\
 ");
       printf ("\
@@ -624,183 +629,286 @@ main (argc, argv)
      char **argv;
 {
   struct termios mode;
-  enum output_type output_type = changed;
+  enum output_type output_type;
   int optc;
+  int require_set_attr;
+  int speed_was_set;
+  int verbose_output;
+  int recoverable_output;
+  int k;
 
   program_name = argv[0];
-  opterr = 0;
 
-  while ((optc = getopt_long (argc, argv, "ag", longopts, (int *) 0)) != EOF)
+  parse_long_options (argc, argv, usage);
+
+  output_type = changed;
+  verbose_output = 0;
+  recoverable_output = 0;
+
+  /* Recognize the long options only.  */
+  opterr = 0;
+  while ((optc = getopt_long_only (argc, argv, "ag", longopts, (int *) 0))
+	 != EOF)
     {
       switch (optc)
 	{
-	case 0:
-	  break;
-
 	case 'a':
+	  verbose_output = 1;
 	  output_type = all;
 	  break;
 
 	case 'g':
+	  recoverable_output = 1;
 	  output_type = recoverable;
 	  break;
 
 	default:
-	  goto done;
+	  break;
 	}
     }
 
-done:;
-
-  if (show_version)
+  /* Recognize short options and combinations: -a, -g, -ag, and -ga.
+     They need not precede non-options.  We cannot use GNU getopt because
+     it would treat -tabs and -ixany as uses of the -a option.  */
+  for (k = optind; k < argc; k++)
     {
-      printf ("%s\n", version_string);
-      exit (0);
+      if (argv[k][0] == '-')
+	{
+	  if (argv[k][1] == 'a'
+	      && argv[k][2] == '\0')
+	    {
+	      ++optind;
+	      verbose_output = 1;
+	    }
+	  else if (argv[k][1] == 'g'
+		   && argv[k][2] == '\0')
+	    {
+	      ++optind;
+	      recoverable_output = 1;
+	    }
+	  else if ((argv[k][1] == 'g'
+	            && argv[k][2] == 'a'
+	            && argv[k][3] == '\0')
+		   || (argv[k][1] == 'a'
+		       && argv[k][2] == 'g'
+		       && argv[k][3] == '\0'))
+	    {
+	      ++optind;
+	      verbose_output = 1;
+	      recoverable_output = 1;
+	    }
+	}
     }
 
-  if (show_help)
-    usage (0);
+  /* Specifying both -a and -g gets an error.  */
+  if (verbose_output && recoverable_output)
+    error (2, 0,
+	   "the options for verbose and stty-readable output styles are\n\
+\tmutually exclusive");
 
+  /* Specifying any other arguments with -a or -g gets an error.  */
+  if (argc - optind > 0 && (verbose_output || recoverable_output))
+    error (2, 0, "when specifying an output style, modes may not be set");
+
+  /* Initialize to all zeroes so there is no risk memcmp will report a
+     spurious difference in an uninitialized portion of the structure.  */
+  bzero (&mode, sizeof (mode));
   if (tcgetattr (0, &mode))
     error (1, errno, "standard input");
 
-  max_col = screen_columns ();
-  current_col = 0;
-
-  if (optind == argc)
+  if (verbose_output || recoverable_output || argc == 1)
     {
-      if (optc == '?')
-	{
-	  error (0, 0, "invalid argument `%s'", argv[--optind]);
-	  usage (1);
-	}
+      max_col = screen_columns ();
+      current_col = 0;
       display_settings (output_type, &mode);
       exit (0);
     }
 
-  while (optind < argc)
+  speed_was_set = 0;
+  require_set_attr = 0;
+  k = 1;
+  while (k < argc)
     {
       int match_found = 0;
       int reversed = 0;
       int i;
 
-      if (argv[optind][0] == '-')
+      if (argv[k][0] == '-')
 	{
-	  ++argv[optind];
+	  ++argv[k];
 	  reversed = 1;
 	}
       for (i = 0; mode_info[i].name != NULL; ++i)
 	{
-	  if (!strcmp (argv[optind], mode_info[i].name))
+	  if (!strcmp (argv[k], mode_info[i].name))
 	    {
 	      match_found = set_mode (&mode_info[i], reversed, &mode);
+	      require_set_attr = 1;
 	      break;
 	    }
 	}
       if (match_found == 0 && reversed)
 	{
-	  error (0, 0, "invalid argument `%s'", --argv[optind]);
+	  error (0, 0, "invalid argument `%s'", --argv[k]);
 	  usage (1);
 	}
       if (match_found == 0)
 	{
 	  for (i = 0; control_info[i].name != NULL; ++i)
 	    {
-	      if (!strcmp (argv[optind], control_info[i].name))
+	      if (!strcmp (argv[k], control_info[i].name))
 		{
-		  if (optind == argc - 1)
+		  if (k == argc - 1)
 		    {
-		      error (0, 0, "missing argument to `%s'", argv[optind]);
+		      error (0, 0, "missing argument to `%s'", argv[k]);
 		      usage (1);
 		    }
 		  match_found = 1;
-		  ++optind;
-		  set_control_char (&control_info[i], argv[optind], &mode);
+		  ++k;
+		  set_control_char (&control_info[i], argv[k], &mode);
+		  require_set_attr = 1;
 		  break;
 		}
 	    }
 	}
       if (match_found == 0)
 	{
-	  if (!strcmp (argv[optind], "ispeed"))
+	  if (!strcmp (argv[k], "ispeed"))
 	    {
-	      if (optind == argc - 1)
+	      if (k == argc - 1)
 		{
-		  error (0, 0, "missing argument to `%s'", argv[optind]);
+		  error (0, 0, "missing argument to `%s'", argv[k]);
 		  usage (1);
 		}
-	      ++optind;
-	      set_speed (input_speed, argv[optind], &mode);
+	      ++k;
+	      set_speed (input_speed, argv[k], &mode);
+	      speed_was_set = 1;
+	      require_set_attr = 1;
 	    }
-	  else if (!strcmp (argv[optind], "ospeed"))
+	  else if (!strcmp (argv[k], "ospeed"))
 	    {
-	      if (optind == argc - 1)
+	      if (k == argc - 1)
 		{
-		  error (0, 0, "missing argument to `%s'", argv[optind]);
+		  error (0, 0, "missing argument to `%s'", argv[k]);
 		  usage (1);
 		}
-	      ++optind;
-	      set_speed (output_speed, argv[optind], &mode);
+	      ++k;
+	      set_speed (output_speed, argv[k], &mode);
+	      speed_was_set = 1;
+	      require_set_attr = 1;
 	    }
 #ifdef TIOCGWINSZ
-	  else if (!strcmp (argv[optind], "rows"))
+	  else if (!strcmp (argv[k], "rows"))
 	    {
-	      if (optind == argc - 1)
+	      if (k == argc - 1)
 		{
-		  error (0, 0, "missing argument to `%s'", argv[optind]);
+		  error (0, 0, "missing argument to `%s'", argv[k]);
 		  usage (1);
 		}
-	      ++optind;
-	      set_window_size ((int) integer_arg (argv[optind]), -1);
+	      ++k;
+	      set_window_size ((int) integer_arg (argv[k]), -1);
 	    }
-	  else if (!strcmp (argv[optind], "cols")
-		   || !strcmp (argv[optind], "columns"))
+	  else if (!strcmp (argv[k], "cols")
+		   || !strcmp (argv[k], "columns"))
 	    {
-	      if (optind == argc - 1)
+	      if (k == argc - 1)
 		{
-		  error (0, 0, "missing argument to `%s'", argv[optind]);
+		  error (0, 0, "missing argument to `%s'", argv[k]);
 		  usage (1);
 		}
-	      ++optind;
-	      set_window_size (-1, (int) integer_arg (argv[optind]));
+	      ++k;
+	      set_window_size (-1, (int) integer_arg (argv[k]));
 	    }
-	  else if (!strcmp (argv[optind], "size"))
+	  else if (!strcmp (argv[k], "size"))
 	    display_window_size (0);
 #endif
 #ifdef HAVE_C_LINE
-	  else if (!strcmp (argv[optind], "line"))
+	  else if (!strcmp (argv[k], "line"))
 	    {
-	      if (optind == argc - 1)
+	      if (k == argc - 1)
 		{
-		  error (0, 0, "missing argument to `%s'", argv[optind]);
+		  error (0, 0, "missing argument to `%s'", argv[k]);
 		  usage (1);
 		}
-	      ++optind;
-	      mode.c_line = integer_arg (argv[optind]);
+	      ++k;
+	      mode.c_line = integer_arg (argv[k]);
+	      require_set_attr = 1;
 	    }
 #endif
-	  else if (!strcmp (argv[optind], "speed"))
+	  else if (!strcmp (argv[k], "speed"))
 	    display_speed (&mode, 0);
-	  else if (string_to_baud (argv[optind]) != (speed_t) -1)
-	    set_speed (both_speeds, argv[optind], &mode);
-	  else if (recover_mode (argv[optind], &mode) == 0)
+	  else if (string_to_baud (argv[k]) != (speed_t) -1)
 	    {
-	      error (0, 0, "invalid argument `%s'", argv[optind]);
-	      usage (1);
+	      set_speed (both_speeds, argv[k], &mode);
+	      speed_was_set = 1;
+	      require_set_attr = 1;
+	    }
+	  else
+	    {
+	      if (recover_mode (argv[k], &mode) == 0)
+		{
+		  error (0, 0, "invalid argument `%s'", argv[k]);
+		  usage (1);
+		}
+	      require_set_attr = 1;
 	    }
 	}
-      optind++;
+      k++;
     }
 
-  if (tcsetattr (0, TCSADRAIN, &mode))
-    error (1, errno, "standard input");
+  if (require_set_attr)
+    {
+      struct termios new_mode;
+
+      if (tcsetattr (0, TCSADRAIN, &mode))
+	error (1, errno, "standard input");
+
+      /* POSIX (according to Zlotnick's book) tcsetattr returns zero if
+	 it performs *any* of the requested operations.  This means it
+	 can report `success' when it has actually failed to perform
+	 some proper subset of the requested operations.  To detect
+	 this partial failure, get the current terminal attributes and
+	 compare them to the requested ones.  */
+
+      /* Initialize to all zeroes so there is no risk memcmp will report a
+	 spurious difference in an uninitialized portion of the structure.  */
+      bzero (&new_mode, sizeof (new_mode));
+      if (tcgetattr (0, &new_mode))
+	error (1, errno, "standard input");
+
+      /* Normally, one shouldn't use memcmp to compare structures that
+	 may have `holes' containing uninitialized data, but we have been
+	 careful to initialize the storage of these two variables to all
+	 zeroes.  One might think it more efficient simply to compare the
+	 modified fields, but that would require enumerating those fields --
+	 and not all systems have the same fields in this structure.  */
+
+      if (memcmp (&mode, &new_mode, sizeof (mode)) != 0)
+	{
+#ifdef CIBAUD
+	  /* SunOS 4.1.3 (at least) has the problem that after this sequence,
+	       tcgetattr(&m1); tcsetattr(&m1); tcgetattr(&m2); 
+	     sometimes (m1 != m2).  The only difference is in the four bits
+	     of the c_cflag field corresponding to the baud rate.  To save
+	     Sun users a little confusion, don't report an error if this
+	     happens.  But suppress the error only if we haven't tried to
+	     set the baud rate explicitly -- otherwise we'd never give an
+	     error for a true failure to set the baud rate.  */
+
+          new_mode.c_cflag &= (~CIBAUD);
+	  if (speed_was_set || memcmp (&mode, &new_mode, sizeof (mode)) != 0)
+#endif
+	  error (1, 0,
+	       "standard input: unable to perform all requested operations");
+	}
+    }
 
   exit (0);
 }
 
 /* Return 0 if not applied because not reversible; otherwise return 1. */
 
-int
+static int
 set_mode (info, reversed, mode)
      struct mode_info *info;
      int reversed;
@@ -999,7 +1107,7 @@ set_mode (info, reversed, mode)
   return 1;
 }
 
-void
+static void
 set_control_char (info, arg, mode)
      struct control_info *info;
      char *arg;
@@ -1025,7 +1133,7 @@ set_control_char (info, arg, mode)
   mode->c_cc[info->offset] = value;
 }
 
-void
+static void
 set_speed (type, arg, mode)
      enum speed_setting type;
      char *arg;
@@ -1041,45 +1149,120 @@ set_speed (type, arg, mode)
 }
 
 #ifdef TIOCGWINSZ
-void
+
+/* Get window size information.  First try getting the information
+   associated with standard output and if that fails, try standard input.
+   Return zero for success, non-zero if both ioctl's failed.  */
+
+static int
+get_win_size (win)
+     struct winsize *win;
+{
+  int err;
+
+  err = ioctl (1, TIOCGWINSZ, (char *) win);
+  if (err != 0)
+    err = ioctl (0, TIOCGWINSZ, (char *) win);
+  return err;
+}
+
+static void
 set_window_size (rows, cols)
      int rows, cols;
 {
   struct winsize win;
 
-  if (ioctl (0, TIOCGWINSZ, (char *) &win))
-    error (1, errno, "standard input");
-  if (rows >= 0)
-    win.ws_row = rows;
-  if (cols >= 0)
-    win.ws_col = cols;
+  if (get_win_size (&win))
+    {
+      if (errno != EINVAL)
+	error (1, errno, "standard input");
+      bzero (&win, sizeof (win));
+    }
+      if (rows >= 0)
+	win.ws_row = rows;
+      if (cols >= 0)
+	win.ws_col = cols;
+
+#ifdef TIOCSSIZE
+  /* Alexander Dupuy <dupuy@cs.columbia.edu> wrote:
+     The following code deals with a bug in the SunOS 4.x (and 3.x?) kernel.
+     This comment from sys/ttold.h describes Sun's twisted logic - a better
+     test would have been (ts_lines > 64k || ts_cols > 64k || ts_cols == 0).
+     At any rate, the problem is gone in Solaris 2.x.
+
+       Unfortunately, the old TIOCSSIZE code does collide with TIOCSWINSZ,
+       but they can be disambiguated by checking whether a "struct ttysize"
+       structure's "ts_lines" field is greater than 64K or not.  If so,
+       it's almost certainly a "struct winsize" instead.
+
+     At any rate, the bug manifests itself when ws_row == 0; the symptom is
+     that ws_row is set to ws_col, and ws_col is set to (ws_xpixel<<16) +
+     ws_ypixel.  Since GNU stty sets rows and columns separately, this bug
+     caused "stty rows 0 cols 0" to set rows to cols and cols to 0, while
+     "stty cols 0 rows 0" would do the right thing.  On a little-endian
+     machine like the sun386i, the problem is the same, but for ws_col == 0.
+
+     The workaround is to do the ioctl once with row and col = 1 to set the
+     pixel info, and then do it again using a TIOCSSIZE to set rows/cols.  */
+
+  if (win.ws_row == 0 || win.ws_col == 0)
+    {
+      struct ttysize ttysz;
+
+      ttysz.ts_lines = win.ws_row;
+      ttysz.ts_cols = win.ws_col;
+
+      win.ws_row = 1;
+      win.ws_col = 1;
+
+      if (ioctl (0, TIOCSWINSZ, (char *) &win))
+	error (1, errno, "standard input");
+
+      if (ioctl (0, TIOCSSIZE, (char *) &ttysz))
+	error (1, errno, "standard input");
+      return;
+    }
+#endif
+
   if (ioctl (0, TIOCSWINSZ, (char *) &win))
     error (1, errno, "standard input");
 }
 
-void
+static void
 display_window_size (fancy)
      int fancy;
 {
   struct winsize win;
 
-  if (ioctl (0, TIOCGWINSZ, (char *) &win))
-    error (1, errno, "standard input");
-  wrapf (fancy ? "rows %d; columns %d;" : "%d %d\n", win.ws_row, win.ws_col);
-  if (!fancy)
-    current_col = 0;
+  if (get_win_size (&win))
+    {
+      if (errno != EINVAL)
+	error (1, errno, "standard input");
+    }
+  else
+    {
+      wrapf (fancy ? "rows %d; columns %d;" : "%d %d\n",
+	     win.ws_row, win.ws_col);
+      if (!fancy)
+	current_col = 0;
+    }
 }
 #endif
 
-int
+static int
 screen_columns ()
 {
 #ifdef TIOCGWINSZ
   struct winsize win;
 
-  if (ioctl (0, TIOCGWINSZ, (char *) &win))
-    error (1, errno, "standard input");
-  if (win.ws_col > 0)
+  if (get_win_size (&win))
+    {
+      /* With Solaris 2.[123], this ioctl fails and errno is set to
+	 EINVAL for telnet (but not rlogin) sessions.  */
+      if (errno != EINVAL)
+	error (1, errno, "standard input");
+    }
+  else if (win.ws_col > 0)
     return win.ws_col;
 #endif
   if (getenv ("COLUMNS"))
@@ -1087,7 +1270,7 @@ screen_columns ()
   return 80;
 }
 
-tcflag_t *
+static tcflag_t *
 mode_type_flag (type, mode)
      enum mode_type type;
      struct termios *mode;
@@ -1114,7 +1297,7 @@ mode_type_flag (type, mode)
     }
 }
 
-void
+static void
 display_settings (output_type, mode)
      enum output_type output_type;
      struct termios *mode;
@@ -1135,7 +1318,7 @@ display_settings (output_type, mode)
     }
 }
 
-void
+static void
 display_changed (mode)
      struct termios *mode;
 {
@@ -1207,7 +1390,7 @@ display_changed (mode)
   current_col = 0;
 }
 
-void
+static void
 display_all (mode)
      struct termios *mode;
 {
@@ -1256,7 +1439,7 @@ display_all (mode)
   current_col = 0;
 }
 
-void
+static void
 display_speed (mode, fancy)
      struct termios *mode;
      int fancy;
@@ -1272,7 +1455,7 @@ display_speed (mode, fancy)
     current_col = 0;
 }
 
-void
+static void
 display_recoverable (mode)
      struct termios *mode;
 {
@@ -1286,7 +1469,7 @@ display_recoverable (mode)
   putchar ('\n');
 }
 
-int
+static int
 recover_mode (arg, mode)
      char *arg;
      struct termios *mode;
@@ -1352,7 +1535,7 @@ struct speed_map speeds[] =
   {NULL, 0, 0}
 };
 
-speed_t
+static speed_t
 string_to_baud (arg)
      char *arg;
 {
@@ -1364,7 +1547,7 @@ string_to_baud (arg)
   return (speed_t) -1;
 }
 
-unsigned long
+static unsigned long
 baud_to_value (speed)
      speed_t speed;
 {
@@ -1376,7 +1559,7 @@ baud_to_value (speed)
   return 0;
 }
 
-void
+static void
 sane_mode (mode)
      struct termios *mode;
 {
@@ -1410,7 +1593,7 @@ sane_mode (mode)
 /* Return a string that is the printable representation of character CH.  */
 /* Adapted from `cat' by Torbjorn Granlund.  */
 
-char *
+static char *
 visible (ch)
      unsigned char ch;
 {
@@ -1463,7 +1646,7 @@ visible (ch)
    but allowing octal and hex numbers as in C.  */
 /* From `od' by Richard Stallman.  */
 
-long
+static long
 integer_arg (s)
      char *s;
 {
